#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <cstring>
#include <iostream>
#include "cuda.h"

#include "helper_multiprocess.h"

// includes, project
#include <helper_functions.h>
#include "helper_cuda_drvapi.h"

// includes, CUDA
#include <builtin_types.h>

using namespace std;

#ifndef UNIT_TEST_PRINTS
// Set this to 1 to verify the correctness of the host-computed matrix.
#define UNIT_TEST_PRINTS 1
#endif

// For direct NVLINK and PCI-E peers, at max 8 simultaneous peers are allowed
// For NVSWITCH connected peers like DGX-2, simultaneous peers are not limited
// in the same way.
#define MAX_DEVICES (32)

#define PROCESSES_PER_DEVICE 1


typedef struct shmStruct_st 
{
  size_t nprocesses;
  int tapID_device_0;
  int tapID_device_1;
  int dims_x;
  int dims_y;
  int dims_z;
  int frame_size;
  int sample_size;
} shmStruct;


// `ipcHandleTypeFlag` specifies the platform specific handle type this sample
// uses for importing and exporting memory allocation. On Linux this sample
// specifies the type as CU_MEM_HANDLE_TYPE_POSIX_FILE_DESCRIPTOR meaning that
// file descriptors will be used. On Windows this sample specifies the type as
// CU_MEM_HANDLE_TYPE_WIN32 meaning that NT HANDLEs will be used. The
// ipcHandleTypeFlag variable is a convenience variable and is passed by value
// to individual requests.
#if defined(__linux__)
CUmemAllocationHandleType ipcHandleTypeFlag = CU_MEM_HANDLE_TYPE_POSIX_FILE_DESCRIPTOR;
#else
CUmemAllocationHandleType ipcHandleTypeFlag = CU_MEM_HANDLE_TYPE_WIN32;
#endif

#if defined(__linux__)
#define cpu_atomic_add32(a, x) __sync_add_and_fetch(a, x)
#elif defined(WIN32) || defined(_WIN32) || defined(WIN64) || defined(_WIN64)
#define cpu_atomic_add32(a, x) InterlockedAdd((volatile LONG *)a, x)
#else
#error Unsupported system
#endif


__global__ void printKernel(float *A, dim3 dimsA, int nCols, int nRows)
{
    // dimsA.x = number of columns
    // dimsA.y = number of rows
    // dimsA.z = number of values per sample, xyz=3 
    int size = dimsA.x * dimsA.y * dimsA.z;
    printf("\n\n[\n");
    
    for (int i=0; i<nRows; i++)
    {
        for (int j=0; j<nCols*dimsA.z; j+= dimsA.z)
        {
            int index = (i*dimsA.x*dimsA.z) + j;

            if (index < size)
            {
                printf("(");
                for (int m = 0; m < dimsA.z; m++ )
                {
                    if (m+1 >= dimsA.z)
                        printf("%6.0f",(float)A[index+m]);
                    else
                        printf("%6.0f,",(float)A[index+m]);
                }
                printf(")\t\t");
            }
        }
        printf("\n");
    }
    printf("]\n\n");
}

// Windows-specific LPSECURITYATTRIBUTES
void getDefaultSecurityDescriptor(CUmemAllocationProp *prop) 
{
#if defined(__linux__)
  return;
#elif defined(WIN32) || defined(_WIN32) || defined(WIN64) || defined(_WIN64)
  static const char sddl[] = "D:P(OA;;GARCSDWDWOCCDCLCSWLODTWPRPCRFA;;;WD)";
  static OBJECT_ATTRIBUTES objAttributes;
  static bool objAttributesConfigured = false;

  if (!objAttributesConfigured) 
  {
    PSECURITY_DESCRIPTOR secDesc;
    BOOL result = ConvertStringSecurityDescriptorToSecurityDescriptorA(sddl, SDDL_REVISION_1, &secDesc, NULL);
    
    if (result == 0) 
    {
      printf("IPC failure: getDefaultSecurityDescriptor Failed! (%d)\n", GetLastError());
    }

    InitializeObjectAttributes(&objAttributes, NULL, 0, NULL, secDesc);

    objAttributesConfigured = true;
  }

  prop->win32HandleMetaData = &objAttributes;
  return;
#endif
}


static void memMapImportAndMapMemory(CUdeviceptr d_ptr, size_t mapSize, std::vector<ShareableHandle> &shareableHandles, int mapDevice) 
{
  std::vector<CUmemGenericAllocationHandle> allocationHandles;
  allocationHandles.resize(shareableHandles.size());

  // The accessDescriptor will describe the mapping requirement for the mapDevice passed as argument
  CUmemAccessDesc accessDescriptor;

  // Specify location for mapping the imported allocations.
  accessDescriptor.location.type = CU_MEM_LOCATION_TYPE_DEVICE;
  accessDescriptor.location.id = mapDevice;

  // Specify both read and write accesses.
  accessDescriptor.flags = CU_MEM_ACCESS_FLAGS_PROT_READWRITE;

  printf("\t [memMapImportAndMapMemory] tap_info shHandles[0]: %d\n", shareableHandles[0] );
  printf("\t [memMapImportAndMapMemory] tap_info shHandles[1]: %d\n", shareableHandles[1] );
  fflush(stdout);


  for (int i = 0; i < shareableHandles.size(); i++) 
  {
    // Import the memory allocation back into a CUDA handle from the platform specific handle.
    printf("\t [memMapImportAndMapMemory Importing] shHandles[i]: %d\n", (void *)(uintptr_t)shareableHandles[i] );
    checkCudaErrors(cuMemImportFromShareableHandle(&allocationHandles[i], (void *)(uintptr_t)shareableHandles[i], ipcHandleTypeFlag));

    // Assign the chunk to the appropriate VA range and release the handle.
    // After mapping the memory, it can be referenced by virtual address.
    checkCudaErrors(cuMemMap(d_ptr + (i * mapSize), mapSize, 0, allocationHandles[i], 0));

    // Since we do not need to make any other mappings of this memory or export
    // it, we no longer need and can release the allocationHandle. The
    // allocation will be kept live until it is unmapped.
    checkCudaErrors(cuMemRelease(allocationHandles[i]));
  }

  // Retain peer access and map all chunks to mapDevice
  checkCudaErrors(cuMemSetAccess(d_ptr, shareableHandles.size() * mapSize, &accessDescriptor, 1));
}

static void memMapUnmapAndFreeMemory(CUdeviceptr dptr, size_t size) 
{
  // Unmap the mapped virtual memory region
  // Since the handles to the mapped backing stores have already been released
  // by cuMemRelease, and these are the only/last mappings referencing them,
  // The backing stores will be freed.
  // Since the memory has been unmapped after this call, accessing the specified
  // va range will result in a fault (unitll it is remapped).
  checkCudaErrors(cuMemUnmap(dptr, size));

  // Free the virtual address region.  This allows the virtual address region
  // to be reused by future cuMemAddressReserve calls.  This also allows the
  // virtual address region to be used by other allocation made through
  // opperating system calls like malloc & mmap.
  checkCudaErrors(cuMemAddressFree(dptr, size));
}

int wuipcRecvShareableHandle(ShareableHandle *shHandle, int tap_handle_0, int tap_handle_1, int i) 
{
    if (i ==0)
    {
      *(int *)shHandle = tap_handle_0;
      //fprintf(stderr," receiv i=0: %d | addr %d", tap_handle_0, &shHandle);
    }
    else
    {
      *(int *)shHandle = tap_handle_1;
      //fprintf(stderr," receiv i=1: %d | addr %d", tap_handle_1, &shHandle);
    }

    return 0;
}

int wuipcRecvShareableHandles(std::vector<ShareableHandle> &shareableHandles, int tap_handle_0, int tap_handle_1) 
{
  for (int i = 0; i < shareableHandles.size(); i++) 
  {
    wuipcRecvShareableHandle(&shareableHandles[i], tap_handle_0, tap_handle_1, i);
  }
  return 0;
}


static void openPCT(CUdeviceptr &d_ptr, CUcontext &ctx, char *tapname, int deviceID, sharedMemoryInfo &info, volatile shmStruct *&shm)
{

  if (sharedMemoryOpen(tapname, sizeof(shmStruct), &info) != 0) 
  {
    printf("Failed to create shared memory slab\n");
    exit(EXIT_FAILURE);
  }
  shm = (volatile shmStruct *)info.addr;
  
  dim3 o_dimsPCT(64, 64, 3);
  o_dimsPCT.x = shm->dims_x;
  o_dimsPCT.y = shm->dims_y;
  o_dimsPCT.z = shm->dims_z;
  int o_frameSize = shm->frame_size;
  int o_sampleSize = shm->sample_size;
  int o_tap_handle_0 = shm->tapID_device_0;
  int o_tap_handle_1 = shm->tapID_device_1;
  size_t procCount = shm->nprocesses;

  printf("Opening Tap: %s ...\n", tapname);
  printf("  - Tap Shared Mem Info -\n");
  printf("\t nprocesses: %ld\n", procCount );
  printf("\t tap_info shHandles[0]: %d\n", o_tap_handle_0 );
  printf("\t tap_info shHandles[1]: %d\n", o_tap_handle_1 );
  printf("\t dims_x: %d\n", o_dimsPCT.x );
  printf("\t dims_y: %d\n", o_dimsPCT.y );
  printf("\t dims_z: %d\n", o_dimsPCT.z );
  printf("\t frame size: %d\n", o_frameSize );
  printf("\t sample size: %d\n", o_sampleSize );

  // Receive all allocation handles shared.
  std::vector<ShareableHandle> o_shHandle(procCount);

  // TODO Only supports 2 GPUs
  wuipcRecvShareableHandles(o_shHandle, o_tap_handle_0, o_tap_handle_1);

  CUdevice device;

  checkCudaErrors(cuDeviceGet(&device, deviceID));
  checkCudaErrors(cuCtxCreate(&ctx, 0, device));
  checkCudaErrors(cuCtxSetCurrent(ctx));


  size_t o_size_bytes = o_dimsPCT.x * o_dimsPCT.y * o_dimsPCT.z * o_sampleSize;

  // Reserve the required contiguous VA space for the allocations
  checkCudaErrors(cuMemAddressReserve(&d_ptr, procCount * o_size_bytes, o_size_bytes, 0, 0));

  // Import the memory allocations shared by the parent with us and map them in our address space.
  memMapImportAndMapMemory(d_ptr, o_size_bytes, o_shHandle, deviceID);

  sharedMemoryClose(&info);

}

// Host code
int main(int argc, char **argv) 
{
  printf("[Consumer] - Starting...\n");

  #if defined(__arm__) || defined(__aarch64__)
    printf("Not supported on ARM\n");
    return EXIT_WAIVED;
  #else
  
    // Initialize
    checkCudaErrors(cuInit(0));
    
    if (checkCmdLineFlag(argc, (const char **)argv, "help") || checkCmdLineFlag(argc, (const char **)argv, "?"))
    {
      printf("Usage -id=tap_name \n");
      printf("      -d=deviceIDToOpenTap \n");
      printf("      -s=secondsToSleepwithSharedTapOpen \n");
      printf(" Example Usage: ./consumer id=\"test-tap-1\" -d=0 -s=30\n");

      exit(EXIT_SUCCESS);
    }

    // program params
    dim3 dimsPCT(64, 64, 3);
    int sampleSize = 3;
    char *tap_name;
    int deviceID = 0;
    typedef float sampletype_t; // declares sampletype_t to be an alias for the type float
    int seconds_to_sleep = 0;
    size_t procCount = 0;

    // Tap Name
    if (checkCmdLineFlag(argc, (const char **)argv, "id"))
    {
      getCmdLineArgumentString(argc, (const char **)argv, "id", &tap_name);
    }

    // Device ID to Open Tap On
    if (checkCmdLineFlag(argc, (const char **)argv, "d"))
    {
      deviceID = getCmdLineArgumentInt(argc, (const char **)argv, "d");
    }

    // Seconds to Sleep
    if (checkCmdLineFlag(argc, (const char **)argv, "s"))
    {
      seconds_to_sleep = getCmdLineArgumentInt(argc, (const char **)argv, "s");
    }

    printf("Opening Tap: %s ...\n\n", tap_name);

    // Tap Objects to unallocate
    sharedMemoryInfo shmInfo;
    volatile shmStruct *shm = NULL;
    CUcontext ctx;
    CUdeviceptr d_ptr = 0ULL;
    
    // Open the Tap
    openPCT(d_ptr, ctx, tap_name, deviceID, shmInfo, shm);

    // some tap meta data
    dimsPCT.x = shm->dims_x;
    dimsPCT.y = shm->dims_y;
    dimsPCT.z = shm->dims_z;
    sampleSize = shm->sample_size;
    procCount = shm->nprocesses;

    // Example Print Kernel
    int nColsToPrint = dimsPCT.x;
    int nRowsToPrint = dimsPCT.y;
    printKernel<<< 1, 1 >>>((float*)d_ptr, dimsPCT, nColsToPrint, nRowsToPrint);
    cudaDeviceSynchronize();

    // Sleep - Keep Tap Open in Memory
    printf("Sleeping for %d seconds...\n", seconds_to_sleep);
    usleep(seconds_to_sleep * 1e6);

    // Cleanup
      printf("Cleaning up...\n");
      checkCudaErrors(cuCtxDestroy(ctx));

      // Unmap the allocations from our address space. Unmapping will also free the
      // handle as we have already released the imported handle with the call to
      // cuMemRelease. Finally, free up the Virtual Address space we reserved with
      // cuMemAddressReserve.
      size_t size_bytes = dimsPCT.x * dimsPCT.y * dimsPCT.z * sampleSize;
      memMapUnmapAndFreeMemory(d_ptr, procCount * size_bytes);

      sharedMemoryClose(&shmInfo);
    
    printf("Done.\n");
    return EXIT_SUCCESS;

  #endif
}
