#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <cstring>
#include <iostream>
#include "cuda.h"

#include "helper_multiprocess.h"

// includes, project
#include <helper_functions.h>
#include "helper_cuda_drvapi.h"

// includes, CUDA
#include <builtin_types.h>

using namespace std;

#ifndef UNIT_TEST_PRINTS
// Set this to 1 to verify the correctness of the host-computed matrix.
#define UNIT_TEST_PRINTS 1
#endif

// For direct NVLINK and PCI-E peers, at max 8 simultaneous peers are allowed
// For NVSWITCH connected peers like DGX-2, simultaneous peers are not limited
// in the same way.
#define MAX_DEVICES (32)

#define PROCESSES_PER_DEVICE 1


typedef struct shmStruct_st 
{
  size_t nprocesses;
  int tapID_device_0;
  int tapID_device_1;
  int dims_x;
  int dims_y;
  int dims_z;
  int frame_size;
  int sample_size;
} shmStruct;


// `ipcHandleTypeFlag` specifies the platform specific handle type this sample
// uses for importing and exporting memory allocation. On Linux this sample
// specifies the type as CU_MEM_HANDLE_TYPE_POSIX_FILE_DESCRIPTOR meaning that
// file descriptors will be used. On Windows this sample specifies the type as
// CU_MEM_HANDLE_TYPE_WIN32 meaning that NT HANDLEs will be used. The
// ipcHandleTypeFlag variable is a convenience variable and is passed by value
// to individual requests.
#if defined(__linux__)
CUmemAllocationHandleType ipcHandleTypeFlag = CU_MEM_HANDLE_TYPE_POSIX_FILE_DESCRIPTOR;
#else
CUmemAllocationHandleType ipcHandleTypeFlag = CU_MEM_HANDLE_TYPE_WIN32;
#endif

#if defined(__linux__)
#define cpu_atomic_add32(a, x) __sync_add_and_fetch(a, x)
#elif defined(WIN32) || defined(_WIN32) || defined(WIN64) || defined(_WIN64)
#define cpu_atomic_add32(a, x) InterlockedAdd((volatile LONG *)a, x)
#else
#error Unsupported system
#endif

// 3D Matrix Initializer Types
enum Init3DType { INIT_MATRIX_RAMP, INIT_MATRIX_RANDOM, INIT_MATRIX_ONES, INIT_MATRIX_IDENTITY, INIT_MATRIX_SENSOR_RAMP };

__global__ void printKernel(float *A, dim3 dimsA, int nCols, int nRows)
{
    // dimsA.x = number of columns
    // dimsA.y = number of rows
    // dimsA.z = number of values per sample, xyz=3 
    int size = dimsA.x * dimsA.y * dimsA.z;
    printf("\n\n[\n");
    
    for (int i=0; i<nRows; i++)
    {
        for (int j=0; j<nCols*dimsA.z; j+= dimsA.z)
        {
            int index = (i*dimsA.x*dimsA.z) + j;

            if (index < size)
            {
                printf("(");
                for (int m = 0; m < dimsA.z; m++ )
                {
                    if (m+1 >= dimsA.z)
                        printf("%6.0f",(float)A[index+m]);
                    else
                        printf("%6.0f,",(float)A[index+m]);
                }
                printf(")\t\t");
            }
        }
        printf("\n");
    }
    printf("]\n\n");
}

template <class T>
__host__ void init_sensor_ramp_host_matrix(T *a, dim3 &dimsA)
{
    // dims.x - number of columns
    // dims.y - number of rows
    // dims.z - number of samples | xyz - 3

    // Fill a column first with 32 channel measures

    // Validate Matrix dim are suitable for tensor operations
    if( (dimsA.x != 128) || (dimsA.y % 32) )
    {
        printf("Error: For sensor emulation, matrix dimensions.x must be 128 and matrix dimensions.y must be a multiple of 32 and matrix dimensions.z must be 3.");
        exit(EXIT_FAILURE);
    }

    int numChannelsPerFrame = 32;
    int numFrames =  dimsA.y / numChannelsPerFrame;
    int numBeamSectors = dimsA.x;

    // strides
    int x_stride = dimsA.z;                             
    int y_stride = dimsA.x * dimsA.z;                   
    int z_stride = 1;                                   
    int frame_stride = y_stride * numChannelsPerFrame;  

    int counter = 0;

    // loop through all frames
    for (int i = 0; i < numFrames; i++) 
    {
        // emulate receiving all beam sectors
        for (int j = 0; j < numBeamSectors; j++)
        {
            // emulate receiving samples for a single beam sector
            for (int k = 0; k < numChannelsPerFrame; k++)
            {
                int offset =  (i * frame_stride) + (j * x_stride) + (k * y_stride);
                
                for (int m = 0; m < dimsA.z; m++)
                {
                    a[offset + m] = (T)counter;
                    counter+=z_stride;
                }  
            } 
        }
        
    }
    
    #if UNIT_TEST_PRINTS 
    {   // Tests Eq. n3 + N3 * (n2 + N2 * n1)
            // x=cols, y=rows, z=depth
            // x=n2,   y=n1,   z=n3
            int n1 = 1; int n2 = 0; int n3=0;
            int offset = n3 + dimsA.z * ( n2 + dimsA.x * n1);
            printf("col: [%d] row: [%d] depth: [%d:] has offset=%d = ",n2,n1,n3,offset);
            printf("(");
            for (int m = 0; m < dimsA.z; m++ )
            {
                if (m+1 >= dimsA.z)
                    printf("%d",(short)a[offset+m]);
                else
                    printf("%d,",(short)a[offset+m]);
            }
            printf(")\n");
            
    }

    {   // Tests Eq. n3 + N3 * (n2 + N2 * n1)
            // x=cols, y=rows, z=depth
            // x=n2,   y=n1,   z=n3
            int n1 = 2; int n2 = 0; int n3=0;
            int offset = n3 + dimsA.z * ( n2 + dimsA.x * n1);
            printf("col: [%d] row: [%d] depth: [%d:] has offset=%d = ",n2,n1,n3,offset);
            printf("(");
            for (int m = 0; m < dimsA.z; m++ )
            {
                if (m+1 >= dimsA.z)
                    printf("%d",(short)a[offset+m]);
                else
                    printf("%d,",(short)a[offset+m]);
            }
            printf(")\n");
    }

    {   // Tests Eq. n3 + N3 * (n2 + N2 * n1)
            // x=cols, y=rows, z=depth
            // x=n2,   y=n1,   z=n3
            int n1 = 2; int n2 = 1; int n3=0;
            int offset = n3 + dimsA.z * ( n2 + dimsA.x * n1);
            printf("col: [%d] row: [%d] depth: [%d:] has offset=%d = ",n2,n1,n3,offset);
            printf("(");
            for (int m = 0; m < dimsA.z; m++ )
            {
                if (m+1 >= dimsA.z)
                    printf("%d",(short)a[offset+m]);
                else
                    printf("%d,",(short)a[offset+m]);
            }
            printf(")\n");
    }

    {   // Tests Eq. n3 + N3 * (n2 + N2 * n1)
            // x=cols, y=rows, z=depth
            // x=n2,   y=n1,   z=n3
            int n1 = 2; int n2 = 2; int n3=0;
            int offset = n3 + dimsA.z * ( n2 + dimsA.x * n1);
            printf("col: [%d] row: [%d] depth: [%d:] has offset=%d = ",n2,n1,n3,offset);
            printf("(");
            for (int m = 0; m < dimsA.z; m++ )
            {
                if (m+1 >= dimsA.z)
                    printf("%d",(short)a[offset+m]);
                else
                    printf("%d,",(short)a[offset+m]);
            }
            printf(")\n");
    }

    {   // Tests Eq. n3 + N3 * (n2 + N2 * n1)
            // x=cols, y=rows, z=depth
            // x=n2,   y=n1,   z=n3
            int n1 = 31; int n2 = 0; int n3=0;
            int offset = n3 + dimsA.z * ( n2 + dimsA.x * n1);
            printf("col: [%d] row: [%d] depth: [%d:] has offset=%d = ",n2,n1,n3,offset);
            printf("(");
            for (int m = 0; m < dimsA.z; m++ )
            {
                if (m+1 >= dimsA.z)
                    printf("%d",(short)a[offset+m]);
                else
                    printf("%d,",(short)a[offset+m]);
            }
            printf(")\n");
    }

    {   // Tests Eq. n3 + N3 * (n2 + N2 * n1)
            // x=cols, y=rows, z=depth
            // x=n2,   y=n1,   z=n3
            int n1 = 32; int n2 = 0; int n3=0;
            int offset = n3 + dimsA.z * ( n2 + dimsA.x * n1);
            printf("col: [%d] row: [%d] depth: [%d:] has offset=%d = ",n2,n1,n3,offset);
            printf("(");
            for (int m = 0; m < dimsA.z; m++ )
            {
                if (m+1 >= dimsA.z)
                    printf("%d",(short)a[offset+m]);
                else
                    printf("%d,",(short)a[offset+m]);
            }
            printf(")\n");
    }

    {   // Tests Eq. n3 + N3 * (n2 + N2 * n1)
            // x=cols, y=rows, z=depth
            // x=n2,   y=n1,   z=n3
            int n1 = 32; int n2 = 1; int n3=0;
            int offset = n3 + dimsA.z * ( n2 + dimsA.x * n1);
            printf("col: [%d] row: [%d] depth: [%d:] has offset=%d = ",n2,n1,n3,offset);
            printf("(");
            for (int m = 0; m < dimsA.z; m++ )
            {
                if (m+1 >= dimsA.z)
                    printf("%d",(short)a[offset+m]);
                else
                    printf("%d,",(short)a[offset+m]);
            }
            printf(")\n");
    }
    #endif
    
}

__host__ int gen_rand(int min, int max)
{
    return min + rand() % (max - min + 1 );
}

template <class T>
__host__ void init_rand_host_matrix(T *a, dim3 &dimsA, int min, int max)
{  
    for (int i = 0; i < dimsA.x* dimsA.y*dimsA.z; i++) 
    {
            a[i] = (T)(gen_rand(min, max));
    }
}

template <class T>
int init_pct(CUdeviceptr &pctData, dim3 &dimsPCT, Init3DType matrix_initializer_type, int min=0, int max=0)
{
    printf("Generating random data...\n");
    int sizePCT = dimsPCT.x * dimsPCT.y * dimsPCT.z;

    T *A_h = NULL;
    A_h = (T*) malloc(sizeof(T) * sizePCT);
    if(A_h== NULL)
    {
	   printf("memory allocation failed\n");
    }

    switch(matrix_initializer_type)
    {
        // case INIT_MATRIX_RAMP : init_ramp_host_matrix(A_h, dimsPCT); break;

        case INIT_MATRIX_RANDOM : init_rand_host_matrix(A_h, dimsPCT, min, max); break;

        // case INIT_MATRIX_ONES : init_ones_host_matrix(A_h, dimsPCT); break;

        // case INIT_MATRIX_IDENTITY : init_identity_host_matrix(A_h, dimsPCT); break;

        case INIT_MATRIX_SENSOR_RAMP : init_sensor_ramp_host_matrix(A_h, dimsPCT); break;

        default : init_rand_host_matrix(A_h, dimsPCT, min, max);
    }
        
   
    // copy host memory to device
    CUstream stream;
    checkCudaErrors(cuStreamCreate(&stream, CU_STREAM_NON_BLOCKING));
    checkCudaErrors(cuMemcpyHtoDAsync(pctData, A_h, sizeof(T) * sizePCT, stream));
    checkCudaErrors(cuStreamSynchronize(stream));
    
    
    // clean up host memory
    checkCudaErrors(cuStreamDestroy(stream));
    free(A_h);
    printf("Generation of data complete.\n\n");
    return 1;
}

// Windows-specific LPSECURITYATTRIBUTES
void getDefaultSecurityDescriptor(CUmemAllocationProp *prop) 
{
#if defined(__linux__)
  return;
#elif defined(WIN32) || defined(_WIN32) || defined(WIN64) || defined(_WIN64)
  static const char sddl[] = "D:P(OA;;GARCSDWDWOCCDCLCSWLODTWPRPCRFA;;;WD)";
  static OBJECT_ATTRIBUTES objAttributes;
  static bool objAttributesConfigured = false;

  if (!objAttributesConfigured) 
  {
    PSECURITY_DESCRIPTOR secDesc;
    BOOL result = ConvertStringSecurityDescriptorToSecurityDescriptorA(sddl, SDDL_REVISION_1, &secDesc, NULL);
    
    if (result == 0) 
    {
      printf("IPC failure: getDefaultSecurityDescriptor Failed! (%d)\n", GetLastError());
    }

    InitializeObjectAttributes(&objAttributes, NULL, 0, NULL, secDesc);

    objAttributesConfigured = true;
  }

  prop->win32HandleMetaData = &objAttributes;
  return;
#endif
}

static void memMapAllocateAndExportMemory(unsigned char backingDevice, size_t allocSize, std::vector<CUmemGenericAllocationHandle> &allocationHandles, std::vector<ShareableHandle> &shareableHandles) 
{
  // This property structure describes the physical location where the memory
  // will be allocated via cuMemCreate along with additional properties.
  CUmemAllocationProp prop = {};

  // The allocations will be device pinned memory backed on backingDevice and
  // exportable with the specified handle type.
  prop.type = CU_MEM_ALLOCATION_TYPE_PINNED;
  prop.location.type = CU_MEM_LOCATION_TYPE_DEVICE;

  // Back all allocations on backingDevice.
  prop.location.id = (int)backingDevice;

  // Passing a requestedHandleTypes indicates intention to export this
  // allocation to a platform-specific handle. This sample requests a file
  // descriptor on Linux and NT Handle on Windows.
  prop.requestedHandleTypes = ipcHandleTypeFlag;

  // Get the minimum granularity supported for allocation with cuMemCreate()
  size_t granularity = 0;
  checkCudaErrors(cuMemGetAllocationGranularity(&granularity, &prop, CU_MEM_ALLOC_GRANULARITY_MINIMUM));
  if (allocSize % granularity) 
  {
    printf("Allocation size is not a multiple of minimum supported granularity for this device. Exiting...\n");
    printf("Granularity = %lu\n", granularity);
    printf("allocSize = %lu\n", allocSize);
    exit(EXIT_FAILURE);
  }

  // Windows-specific LPSECURITYATTRIBUTES is required when
  // CU_MEM_HANDLE_TYPE_WIN32 is used. The security attribute defines the scope
  // of which exported allocations may be tranferred to other processes. For all
  // other handle types, pass NULL.
  getDefaultSecurityDescriptor(&prop);

  for (int i = 0; i < allocationHandles.size(); i++) 
  {
    // Create the allocation as a pinned allocation on device specified in
    // prop.location.id
    checkCudaErrors(cuMemCreate(&allocationHandles[i], allocSize, &prop, 0));

    // Export the allocation to a platform-specific handle. The type of handle
    // requested here must match the requestedHandleTypes field in the prop
    // structure passed to cuMemCreate.
    checkCudaErrors(cuMemExportToShareableHandle((void *)&shareableHandles[i], allocationHandles[i], ipcHandleTypeFlag, 0));
  }
}

static void memMapImportAndMapMemory(CUdeviceptr d_ptr, size_t mapSize, std::vector<ShareableHandle> &shareableHandles, int mapDevice) 
{
  std::vector<CUmemGenericAllocationHandle> allocationHandles;
  allocationHandles.resize(shareableHandles.size());

  // The accessDescriptor will describe the mapping requirement for the mapDevice passed as argument
  CUmemAccessDesc accessDescriptor;

  // Specify location for mapping the imported allocations.
  accessDescriptor.location.type = CU_MEM_LOCATION_TYPE_DEVICE;
  accessDescriptor.location.id = mapDevice;

  // Specify both read and write accesses.
  accessDescriptor.flags = CU_MEM_ACCESS_FLAGS_PROT_READWRITE;

  printf("\t [memMapImportAndMapMemory] tap_info shHandles[0]: %d\n", shareableHandles[0] );
  printf("\t [memMapImportAndMapMemory] tap_info shHandles[1]: %d\n", shareableHandles[1] );
  fflush(stdout);

  for (int i = 0; i < shareableHandles.size(); i++) 
  {
    // Import the memory allocation back into a CUDA handle from the platform specific handle.
    printf("\t [memMapImportAndMapMemory Importing] shHandles[i]: %d\n", shareableHandles[i] );
    checkCudaErrors(cuMemImportFromShareableHandle(&allocationHandles[i], (void *)(uintptr_t)shareableHandles[i], ipcHandleTypeFlag));

    // Assign the chunk to the appropriate VA range and release the handle.
    // After mapping the memory, it can be referenced by virtual address.
    checkCudaErrors(cuMemMap(d_ptr + (i * mapSize), mapSize, 0, allocationHandles[i], 0));

    // Since we do not need to make any other mappings of this memory or export
    // it, we no longer need and can release the allocationHandle. The
    // allocation will be kept live until it is unmapped.
    //checkCudaErrors(cuMemRelease(allocationHandles[i]));
  }

  // Retain peer access and map all chunks to mapDevice
  checkCudaErrors(cuMemSetAccess(d_ptr, shareableHandles.size() * mapSize, &accessDescriptor, 1));
}

static void memMapUnmapAndFreeMemory(CUdeviceptr dptr, size_t size) 
{
  // Unmap the mapped virtual memory region
  // Since the handles to the mapped backing stores have already been released
  // by cuMemRelease, and these are the only/last mappings referencing them,
  // The backing stores will be freed.
  // Since the memory has been unmapped after this call, accessing the specified
  // va range will result in a fault (unitll it is remapped).
  checkCudaErrors(cuMemUnmap(dptr, size));

  // Free the virtual address region.  This allows the virtual address region
  // to be reused by future cuMemAddressReserve calls.  This also allows the
  // virtual address region to be used by other allocation made through
  // opperating system calls like malloc & mmap.
  checkCudaErrors(cuMemAddressFree(dptr, size));
}

static void openPCT(CUdeviceptr &d_ptr, CUcontext &ctx, char *tapname, int deviceID)
{
  volatile shmStruct *shm = NULL;
  sharedMemoryInfo info;

  if (sharedMemoryOpen(tapname, sizeof(shmStruct), &info) != 0) 
  {
    printf("Failed to create shared memory slab\n");
    exit(EXIT_FAILURE);
  }
  shm = (volatile shmStruct *)info.addr;
  
  dim3 o_dimsPCT(64, 64, 3);
  o_dimsPCT.x = shm->dims_x;
  o_dimsPCT.y = shm->dims_y;
  o_dimsPCT.z = shm->dims_z;
  int o_frameSize = shm->frame_size;
  int o_sampleSize = shm->sample_size;
  int o_tap_handle_0 = shm->tapID_device_0;
  int o_tap_handle_1 = shm->tapID_device_1;
  size_t procCount = shm->nprocesses;

  printf("Opening Tap: %s ...\n", tapname);
  printf("  - Tap Shared Mem Info -\n");
  printf("\t nprocesses: %ld\n", procCount );
  printf("\t tap_info shHandles[0]: %d\n", o_tap_handle_0 );
  printf("\t tap_info shHandles[1]: %d\n", o_tap_handle_1 );
  printf("\t dims_x: %d\n", o_dimsPCT.x );
  printf("\t dims_y: %d\n", o_dimsPCT.y );
  printf("\t dims_z: %d\n", o_dimsPCT.z );
  printf("\t frame size: %d\n", o_frameSize );
  printf("\t sample size: %d\n", o_sampleSize );

  // Receive all allocation handles shared.
  std::vector<ShareableHandle> o_shHandle(procCount);
  
  // TODO Only supports 2 GPUs
  o_shHandle[0] = (int)o_tap_handle_0;
  o_shHandle[1] = (int)o_tap_handle_1;

  CUdevice device;

  checkCudaErrors(cuDeviceGet(&device, deviceID));
  checkCudaErrors(cuCtxCreate(&ctx, 0, device));

  size_t o_size_bytes = o_dimsPCT.x * o_dimsPCT.y * o_dimsPCT.z * o_sampleSize;

  // Reserve the required contiguous VA space for the allocations
  checkCudaErrors(cuMemAddressReserve(&d_ptr, procCount * o_size_bytes, o_size_bytes, 0, 0));

  // Import the memory allocations shared by the parent with us and map them in our address space.
  memMapImportAndMapMemory(d_ptr, o_size_bytes, o_shHandle, deviceID);

  sharedMemoryClose(&info);

}


template <class T>
static void createPCT(char *tap_name, dim3 &dimsPCT, int frameSize, sharedMemoryInfo &info, volatile shmStruct *&shm, std::vector<ShareableHandle> &shHandles, std::vector<CUmemGenericAllocationHandle> &allocationHandles) 
{
  int devCount, i, nprocesses = 0;
  //volatile shmStruct *shm = NULL;

  checkCudaErrors(cuDeviceGetCount(&devCount));
  std::vector<CUdevice> devices(devCount);

  if (sharedMemoryCreate(tap_name, sizeof(*shm), &info) != 0) 
  {
    printf("Failed to create host shared memory slab\n");
    exit(EXIT_FAILURE);
  }

  shm = (volatile shmStruct *)info.addr;
  memset((void *)shm, 0, sizeof(*shm));

  for (i = 0; i < devCount; i++) 
  {
    checkCudaErrors(cuDeviceGet(&devices[i], i));
  }

  std::vector<CUcontext> ctxs;
  std::vector<unsigned char> selectedDevices;

  // Pick all the devices that can access each other's memory for this test
  // Keep in mind that CUDA has minimal support for fork() without a
  // corresponding exec() in the child process, but in this case our
  // spawnProcess will always exec, so no need to worry.
  for (i = 0; i < devCount; i++) 
  {
    bool allPeers = true;
    int deviceComputeMode;
    int deviceSupportsIpcHandle;
    int attributeVal = 0;

    checkCudaErrors(cuDeviceGet(&devices[i], i));
    checkCudaErrors(cuDeviceGetAttribute(&deviceComputeMode, CU_DEVICE_ATTRIBUTE_COMPUTE_MODE, devices[i]));
    checkCudaErrors(cuDeviceGetAttribute(&attributeVal, CU_DEVICE_ATTRIBUTE_VIRTUAL_ADDRESS_MANAGEMENT_SUPPORTED,devices[i]));
#if defined(__linux__)
    checkCudaErrors(cuDeviceGetAttribute(
        &deviceSupportsIpcHandle,
        CU_DEVICE_ATTRIBUTE_HANDLE_TYPE_POSIX_FILE_DESCRIPTOR_SUPPORTED,
        devices[i]));
#else
    checkCudaErrors(cuDeviceGetAttribute(&deviceSupportsIpcHandle,CU_DEVICE_ATTRIBUTE_HANDLE_TYPE_WIN32_HANDLE_SUPPORTED, devices[i]));
#endif
    // Check that the selected device supports virtual address management
    if (attributeVal == 0) 
    {
      printf("Device %d doesn't support VIRTUAL ADDRESS MANAGEMENT.\n",devices[i]);
      continue;
    }

    // This sample requires two processes accessing each device, so we need
    // to ensure exclusive or prohibited mode is not set
    if (deviceComputeMode != CU_COMPUTEMODE_DEFAULT) 
    {
      printf("Device %d is in an unsupported compute mode for this sample\n",i);
      continue;
    }

    if (!deviceSupportsIpcHandle) 
    {
      printf("Device %d does not support requested handle type for IPC, skipping...\n",i);
      continue;
    }

    for (int j = 0; j < nprocesses; j++) 
    {
      int canAccessPeerIJ, canAccessPeerJI;
      checkCudaErrors(cuDeviceCanAccessPeer(&canAccessPeerJI, devices[j], devices[i]));
      checkCudaErrors(cuDeviceCanAccessPeer(&canAccessPeerIJ, devices[i], devices[j]));
      if (!canAccessPeerIJ || !canAccessPeerJI) 
      {
        allPeers = false;
        break;
      }
    }
    if (allPeers) 
    {
      CUcontext ctx;
      checkCudaErrors(cuCtxCreate(&ctx, 0, devices[i]));
      ctxs.push_back(ctx);

      // Enable peers here.  This isn't necessary for IPC, but it will
      // setup the peers for the device.  For systems that only allow 8
      // peers per GPU at a time, this acts to remove devices from CanAccessPeer
      for (int j = 0; j < nprocesses; j++) 
      {
        checkCudaErrors(cuCtxSetCurrent(ctxs[i]));
        checkCudaErrors(cuCtxEnablePeerAccess(ctxs[j], 0));
        checkCudaErrors(cuCtxSetCurrent(ctxs[j]));
        checkCudaErrors(cuCtxEnablePeerAccess(ctxs[i], 0));
      }
      selectedDevices.push_back(i);
      nprocesses++;
      if (nprocesses >= MAX_DEVICES) 
      {
        break;
      }
    } 
    else 
    {
      printf("Device %d is not peer capable with some other selected peers, skipping\n",i);
    }
  }

  for (int i = 0; i < ctxs.size(); ++i) 
  {
    checkCudaErrors(cuCtxDestroy(ctxs[i]));
  };

  if (nprocesses == 0) 
  {
    printf("No CUDA devices support IPC\n");
    exit(EXIT_WAIVED);
  }
  shm->nprocesses = nprocesses;

  unsigned char firstSelectedDevice = selectedDevices[0];

  shHandles.resize(nprocesses);
  allocationHandles.resize(nprocesses);

  // Allocate `nprocesses` number of memory chunks and obtain a shareable handle for each allocation.

  size_t size_bytes = dimsPCT.x * dimsPCT.y * dimsPCT.z * sizeof(T);
  memMapAllocateAndExportMemory(firstSelectedDevice, size_bytes, allocationHandles, shHandles);

  // TODO Only supports 2 GPUs
  shm->tapID_device_0 = (int)shHandles[0];
  shm->tapID_device_1 = (int)shHandles[1];

  shm->dims_x = dimsPCT.x;
  shm->dims_y = dimsPCT.y;
  shm->dims_z = dimsPCT.z;
  shm->frame_size = frameSize;
  shm->sample_size = sizeof(T);

}

// Host code
int main(int argc, char **argv) 
{
  printf("[Producer] - Starting...\n");

  #if defined(__arm__) || defined(__aarch64__)
    printf("Not supported on ARM\n");
    return EXIT_WAIVED;
  #else
  
    // Initialize
    checkCudaErrors(cuInit(0));
    
    if (checkCmdLineFlag(argc, (const char **)argv, "help") || checkCmdLineFlag(argc, (const char **)argv, "?"))
    {
      printf("Usage -id=tap_name \n");
      printf("      -nF=NumberOfFrames \n");
      printf("      -nA=NumberOfAzimuths \n");
      printf("      -nP=NumberOfPointsPerFrame \n");
      printf("      -nD=NumberOfSampelsPerPoint \n");
      printf("      -d=deviceIDToOpenTap \n");
      printf("      -s=secondsToSleepwithSharedTapOpen \n");
      printf(" Example Usage: ./producer id=\"test-tap-1\" -nF=32 -nA=128 -nP=4096 -nD=4 -s=30\n");

      exit(EXIT_SUCCESS);
    }

    // program params
    dim3 dimsPCT(64, 64, 3);
    int frameSize = 1;
    char *tap_name;
    int deviceID = 0;
    typedef float sampletype_t; // declares sampletype_t to be an alias for the type float
    int seconds_to_sleep = 0;


    // Tap Name
    if (checkCmdLineFlag(argc, (const char **)argv, "id"))
    {
      getCmdLineArgumentString(argc, (const char **)argv, "id", &tap_name);
    }

    // Frame Length
    if (checkCmdLineFlag(argc, (const char **)argv, "nP"))
    {
      frameSize = getCmdLineArgumentInt(argc, (const char **)argv, "nP");
    }

      // Width of xxx | Number of Columns
    if (checkCmdLineFlag(argc, (const char **)argv, "nA"))
    {
      dimsPCT.x = getCmdLineArgumentInt(argc, (const char **)argv, "nA");
    }

    // Height of xxx | Number of Rows
    if (checkCmdLineFlag(argc, (const char **)argv, "nF"))
    {
      dimsPCT.y = (frameSize / dimsPCT.x) * getCmdLineArgumentInt(argc, (const char **)argv, "nF");
    }

    // Depth of xxx | Number of Samples
    if (checkCmdLineFlag(argc, (const char **)argv, "nD"))
    {
      dimsPCT.z = getCmdLineArgumentInt(argc, (const char **)argv, "nD");
    }

    // Device ID to Open Tap On
    if (checkCmdLineFlag(argc, (const char **)argv, "d"))
    {
      deviceID = getCmdLineArgumentInt(argc, (const char **)argv, "d");
    }

    // Seconds to Sleep
    if (checkCmdLineFlag(argc, (const char **)argv, "s"))
    {
      seconds_to_sleep = getCmdLineArgumentInt(argc, (const char **)argv, "s");
    }

    printf("Tap: %s (%d,%d,%d)\n\n", tap_name, dimsPCT.x, dimsPCT.y, dimsPCT.z);

    // Tap Objects to unallocate
    sharedMemoryInfo shmInfo;
    volatile shmStruct *shm = NULL;
    std::vector<ShareableHandle> shHandles;
    std::vector<CUmemGenericAllocationHandle> allocationHandles;

    // Create the Tap
    createPCT<sampletype_t>(tap_name, dimsPCT, frameSize, shmInfo, shm, shHandles, allocationHandles);

    // Open the Tap
    CUcontext ctx;
    CUdeviceptr d_ptr = 0ULL;
    openPCT(d_ptr, ctx, tap_name, deviceID);

    // Create some data
    init_pct<sampletype_t>(d_ptr, dimsPCT, INIT_MATRIX_SENSOR_RAMP);

    // Example Print Kernel
    // int nColsToPrint = dimsPCT.x;
    // int nRowsToPrint = dimsPCT.y;
    //printKernel<<< 1, 1 >>>((float*)d_ptr, dimsPCT, nColsToPrint, nRowsToPrint);
    cudaDeviceSynchronize();

    // Sleep - Keep Tap Open in Memory
    printf("Sleeping for %d seconds...\n", seconds_to_sleep);
    fprintf(stderr,"Sleeping for %d seconds...\n", seconds_to_sleep);
    usleep(seconds_to_sleep * 1e6);

    // Cleanup
      printf("Cleaning up...\n");
      fprintf(stderr,"Cleaning up...\n");
      checkCudaErrors(cuCtxDestroy(ctx));

      // Close the shareable handles
      for (int i = 0; i < shHandles.size(); i++) 
      {
        checkIpcErrors(ipcCloseShareableHandle(shHandles[i]));
      }

      // Unmap the allocations from our address space. Unmapping will also free the
      // handle as we have already released the imported handle with the call to
      // cuMemRelease. Finally, free up the Virtual Address space we reserved with
      // cuMemAddressReserve.
      size_t size_bytes = dimsPCT.x * dimsPCT.y * dimsPCT.z * sizeof(sampletype_t);
      memMapUnmapAndFreeMemory(d_ptr, shm->nprocesses * size_bytes);

      
      for (int i = 0; i < allocationHandles.size(); i++) 
      {
        checkCudaErrors(cuMemRelease(allocationHandles[i]));
      }

      sharedMemoryClose(&shmInfo);
    
    printf("Done.\n");
    return EXIT_SUCCESS;

  #endif
}
